"""Python implementation of MimicDB
"""


class MimicDB(object):
    def __init__(self, backend=None, namespace=None):
        """Initialze the MimicDB backend with an optional namespace.
        """
        if not backend:
            from .backends.default import Redis
            backend = Redis()

        globals()['backend'] = backend

        if namespace:
            from .backends import tpl
            tpl.set_namespace(namespace)



connection = 'mimicdb'
bucket = 'mimicdb:%s'
key = 'mimicdb:%s:%s'

def set_namespace(namespace=None):
    if namespace:
        globals()['connection'] = 'mimicdb:' + namespace
        globals()['bucket'] = 'mimicdb:' + namespace + ':%s'
        globals()['key'] = 'mimicdb:' + namespace + ':%s:%s'



import os
import random
import string
import hashlib
import logging
import unittest

from boto.s3.connection import S3Connection as BotoS3Connection
from boto.s3.bucket import Bucket as BotoBucket
from boto.s3.key import Key as BotoKey
from boto.exception import S3CreateError, S3ResponseError

import mimicdb
from mimicdb import MimicDB
from mimicdb.s3.connection import S3Connection
from mimicdb.s3.bucket import Bucket
from mimicdb.s3.key import Key
from mimicdb.backends import tpl
from mimicdb.backends.sqlite import SQLite
from mimicdb.backends.memory import Memory
from mimicdb.backends.default import Redis

AWS_ACCESS_KEY_ID = os.environ['AWS_ACCESS_KEY_ID']
AWS_SECRET_ACCESS_KEY = os.environ['AWS_SECRET_ACCESS_KEY']

logging.getLogger('boto').setLevel(logging.CRITICAL)

def random_string(size=6, chars=string.ascii_lowercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


class MimicDBTestCase(unittest.TestCase):
    def setUp(self):
        MimicDB(namespace='tests')

        self.conn = S3Connection(AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
        self.boto_conn = BotoS3Connection(AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)

        self.name = 'mimicdb-tests-' + random_string()
        self.boto_bucket = self.boto_conn.create_bucket(self.name)

        key = BotoKey(self.boto_bucket)
        key.key = 'upload'
        key.set_contents_from_string('upload')

    def tearDown(self):
        for bucket in self.boto_conn.get_all_buckets():
            if bucket.name.startswith(self.name):
                for key in bucket.list():
                    key.delete()

                bucket.delete()

        for key in mimicdb.backend.keys(tpl.connection + '*'):
            mimicdb.backend.delete(key)

    def backend_test(self):
        self.assertIsInstance(mimicdb.backend, mimicdb.backends.Backend)

    def get_all_buckets_test(self):
        buckets = self.conn.get_all_buckets()

        assert buckets == []

    def get_all_buckets_force_test(self):
        buckets = self.conn.get_all_buckets(force=True)

        assert self.name in [bucket.name for bucket in buckets]

    def get_bucket_test(self):
        with self.assertRaises(S3ResponseError):
            self.conn.get_bucket(self.name)

    def get_no_validate_bucket_test(self):
        self.conn.get_bucket(self.name, validate=False)

    def get_bucket_force_test(self):
        self.conn.get_bucket(self.name, force=True)

        assert self.name in mimicdb.backend.smembers(tpl.connection)

    def create_bucket_test(self):
        bucket = self.conn.create_bucket(self.name + '-create')

        assert self.name + '-create' in mimicdb.backend.smembers(tpl.connection)

    def sync_test(self):
        self.conn.sync()

        assert self.name in mimicdb.backend.smembers(tpl.connection)
        assert 'upload' in mimicdb.backend.smembers(tpl.bucket % self.name)

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'upload'))

        assert int(meta['size']) == len('upload')
        assert meta['md5'] == hashlib.md5('upload').hexdigest()

    def delete_bucket_not_empty_test(self):
        self.conn.sync()

        with self.assertRaises(S3ResponseError):
            self.conn.delete_bucket(self.name)

        assert self.name in mimicdb.backend.smembers(tpl.connection)

    def delete_bucket_empty_test(self):
        for key in self.boto_bucket.list():
            key.delete()

        self.conn.sync()
        self.conn.delete_bucket(self.name)

        assert self.name not in mimicdb.backend.smembers(tpl.connection)

    def sync_bucket_test(self):
        self.boto_conn.create_bucket(self.name + '-sync')

        self.conn.sync(self.name)

        assert self.name in mimicdb.backend.smembers(tpl.connection)
        assert self.name + '-sync' not in mimicdb.backend.smembers(tpl.connection)

    def sync_clear_test(self):
        self.conn.sync()

        key = BotoKey(self.boto_bucket)
        key.key = 'upload'
        key.set_contents_from_string('sync_clear')

        self.conn.sync()

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'upload'))

        assert int(meta['size']) == len('sync_clear')
        assert meta['md5'] == hashlib.md5('sync_clear').hexdigest()  

    def sync_bucket_clear_test(self):
        self.conn.sync()

        key = BotoKey(self.boto_bucket)
        key.key = 'upload'
        key.set_contents_from_string('sync_clear')

        self.conn.sync(self.name)

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'upload'))

        assert int(meta['size']) == len('sync_clear')
        assert meta['md5'] == hashlib.md5('sync_clear').hexdigest() 

    def get_unsynced_key_test(self):
        self.conn.sync()

        boto_key = BotoKey(self.boto_bucket)
        boto_key.name = 'get_key'
        boto_key.set_contents_from_string('get_key')

        bucket = self.conn.get_bucket(self.name)
        key = bucket.get_key('get_key')

        assert key is None

    def get_unsynced_key_force_test(self):
        self.conn.sync()

        boto_key = BotoKey(self.boto_bucket)
        boto_key.name = 'get_key'
        boto_key.set_contents_from_string('get_key')

        bucket = self.conn.get_bucket(self.name)
        key = bucket.get_key('get_key', force=True)

        assert key is not None

    def delete_keys_string_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        bucket.delete_keys(['upload'])

        boto_bucket = self.boto_conn.get_bucket(self.name)
        key = boto_bucket.get_key('upload')

        assert key is None
        assert 'upload' not in mimicdb.backend.smembers(tpl.bucket % self.name)

    def delete_keys_key_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        key = bucket.get_key('upload')
        bucket.delete_keys([key])

        boto_bucket = self.boto_conn.get_bucket(self.name)
        key = boto_bucket.get_key('upload')

        assert key is None
        assert 'upload' not in mimicdb.backend.smembers(tpl.bucket % self.name)

    def delete_key_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        bucket.delete_key('upload')

        key = self.boto_bucket.get_key('upload')

        assert key is None
        assert 'upload' not in mimicdb.backend.smembers(tpl.bucket % self.name)

    def iter_test(self):
        self.conn.sync()

        key = BotoKey(self.boto_bucket)
        key.key = 'list'
        key.set_contents_from_string('list')

        bucket = self.conn.get_bucket(self.name)

        assert 'list' not in [key.name for key in bucket]

    def list_test(self):
        self.conn.sync()

        key = BotoKey(self.boto_bucket)
        key.key = 'list'
        key.set_contents_from_string('list')

        bucket = self.conn.get_bucket(self.name)
        keys = bucket.list()

        assert 'list' not in [key.name for key in keys]

    def list_force_test(self):
        self.conn.sync()

        key = BotoKey(self.boto_bucket)
        key.key = 'list'
        key.set_contents_from_string('list')

        bucket = self.conn.get_bucket(self.name)
        keys = list(bucket.list(force=True))

        assert 'list' in [key.name for key in keys]
        assert hashlib.md5('list').hexdigest() in [key.md5 for key in keys]

    def get_all_test(self):
        self.conn.sync()
        bucket = self.conn.get_bucket(self.name)

        assert 'upload' in [key.name for key in bucket.get_all_keys()]

    def get_all_force_test(self):
        self.conn.sync()

        boto_key = BotoKey(self.boto_bucket)
        boto_key.name = 'get_all'
        boto_key.set_contents_from_string('get_all')

        bucket = self.conn.get_bucket(self.name)

        assert 'get_all' in [key.name for key in bucket.get_all_keys(force=True)]

    def bucket_sync_test(self):
        self.conn.sync()

        boto_key = BotoKey(self.boto_bucket)
        boto_key.name = 'sync'
        boto_key.set_contents_from_string('sync')

        bucket = self.conn.get_bucket(self.name)
        bucket.sync()

        key = bucket.get_key('sync')

        assert key.size == len('sync')
        assert key.md5 == hashlib.md5('sync').hexdigest()

    def get_key_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        key = Key(bucket)
        key.key = 'upload'

        assert key.key == 'upload'

    def get_unassigned_key_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        key = Key(bucket)

        assert key.key == None

    def set_key_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)
        key = Key(bucket)
        key.key = 'upload'

        assert key.size == len('upload')
        assert key.md5 == hashlib.md5('upload').hexdigest()  

    def track_upload_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = Key(bucket)
        key.key = 'track'
        key.set_contents_from_string('track')

        assert 'track' in mimicdb.backend.smembers(tpl.bucket % self.name)

    def track_upload_init_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = Key(bucket, 'track')
        key.set_contents_from_string('track')

        assert 'track' in mimicdb.backend.smembers(tpl.bucket % self.name)

    def track_upload_meta_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = Key(bucket)
        key.key = 'track'
        key.set_contents_from_string('track')

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'track'))

        assert int(meta['size']) == len('track')
        assert meta['md5'] == hashlib.md5('track').hexdigest()

    def track_upload_meta_init_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = Key(bucket, 'track')
        key.set_contents_from_string('track')

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'track'))

        assert int(meta['size']) == len('track')
        assert meta['md5'] == hashlib.md5('track').hexdigest()

    def track_download_meta_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = BotoKey(self.boto_bucket)
        key.key = 'download'
        key.set_contents_from_string('download')

        key = Key(bucket)
        key.key = 'download'
        key.get_contents_as_string()

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'download'))

        assert int(meta['size']) == len('download')
        assert meta['md5'] == hashlib.md5('download').hexdigest()

    def track_download_meta_init_test(self):
        self.conn.sync()

        bucket = self.conn.get_bucket(self.name)

        key = BotoKey(self.boto_bucket)
        key.key = 'download'
        key.set_contents_from_string('download')

        key = Key(bucket, 'download')
        key.get_contents_as_string()

        meta = mimicdb.backend.hgetall(tpl.key % (self.name, 'download'))

        assert int(meta['size']) == len('download')
        assert meta['md5'] == hashlib.md5('download').hexdigest()


class SQLiteBackendTestCase(unittest.TestCase):
    def setUp(self):
        self.sqlite = SQLite()

    def init_test(self):
        self.assertIsInstance(self.sqlite, mimicdb.backends.Backend)

    def keys_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-keys')

        assert 'mimicdb-tests' in self.sqlite.keys('mimicdb*')

    def delete_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-delete')
        self.sqlite.delete('mimicdb-tests')

        assert 'mimicdb-tests' not in self.sqlite.keys('mimicdb*')

    def sadd_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-keys')

        assert 'test-keys' in self.sqlite.smembers('mimicdb-tests')

    def srem_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-keys')
        self.sqlite.srem('mimicdb-tests', 'test-keys')

        assert 'test-keys' not in self.sqlite.smembers('mimicdb-tests')

    def sismember_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-keys')

        assert self.sqlite.sismember('mimicdb-tests', 'test-keys')

    def smembers_test(self):
        self.sqlite.sadd('mimicdb-tests', 'test-keys')

        assert 'test-keys' in self.sqlite.smembers('mimicdb-tests')

    def hmset_test(self):
        self.sqlite.hmset('mimicdb-tests', dict(size='3', md5='md5'))
        
        assert self.sqlite.hgetall('mimicdb-tests') == dict(size='3', md5='md5')

    def hgetall_test(self):
        self.sqlite.hmset('mimicdb-tests', dict(size='3', md5='md5'))
        
        assert self.sqlite.hgetall('mimicdb-tests') == dict(size='3', md5='md5')

    def hgetall_none_test(self):
        self.sqlite.hmset('mimicdb-tests', dict(size='3', md5='md5'))
        
        assert self.sqlite.hgetall('mimicdb') == dict()


class MemoryBackendTestCase(unittest.TestCase):
    def setUp(self):
        self.memory = Memory()

    def init_test(self):
        self.assertIsInstance(self.memory, mimicdb.backends.Backend)

    def keys_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')

        assert 'mimicdb-tests' in self.memory.keys('mimicdb*')

    def delete_test(self):
        self.memory.sadd('mimicdb-tests', 'test-delete')
        self.memory.delete('mimicdb-tests')

        assert 'mimicdb-tests' not in self.memory.keys('mimicdb*')

    def sadd_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')
        self.memory.sadd('mimicdb-tests', 'test-keys-2')

        assert 'test-keys' in self.memory.smembers('mimicdb-tests')
        assert 'test-keys-2' in self.memory.smembers('mimicdb-tests')

    def srem_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')
        self.memory.srem('mimicdb-tests', 'test-keys')

        assert 'test-keys' not in self.memory.smembers('mimicdb-tests')

    def sismember_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')

        assert self.memory.sismember('mimicdb-tests', 'test-keys')

    def sisntmember_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')

        assert self.memory.sismember('mimicdb', 'test-keys') == False

    def smembers_test(self):
        self.memory.sadd('mimicdb-tests', 'test-keys')

        assert 'test-keys' in self.memory.smembers('mimicdb-tests')

    def hmset_test(self):
        self.memory.hmset('mimicdb-tests', dict(size='3', md5='md5'))
        
        assert self.memory.hgetall('mimicdb-tests') == dict(size='3', md5='md5')

    def hgetall_test(self):
        self.memory.hmset('mimicdb-tests', dict(size='3', md5='md5'))
        
        assert self.memory.hgetall('mimicdb-tests') == dict(size='3', md5='md5')




import sqlite3

from . import Backend


class SQLite(Backend):
    """SQLite backend. Pass ``sqlite3.connect()`` parameters to ``__init__``.
    If no parameters are passed, ``:memory:`` is chossen as the default
    database.

    :param args \*args, \**kwargs: sqlite3.connect() parameters

    .. code-block:: python

        from mimicdb.backends.sqlite import SQLite

        sqlite = SQLite('mimicdb.db')
    """
    def __init__(self, *args, **kwargs):
        if not args and not kwargs:
            args = [':memory:']

        self._sqlite = sqlite3.connect(*args, **kwargs)

    def keys(self, pattern='*'):
        c = self._sqlite.cursor()
        pattern = pattern.replace('*', '%')
        c.execute('SELECT name FROM sqlite_master WHERE type="%s" AND name LIKE "%s"' % ('table', pattern))

        return [row[0] for row in c.fetchall()]

    def delete(self, *names):
        c = self._sqlite.cursor()

        for name in names:
            c.execute('DROP TABLE IF EXISTS "%s"' % (name,))

        self._sqlite.commit()

    def sadd(self, name, *values):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (member text)' % (name,))

        for value in values:
            c.execute('INSERT INTO "%s" VALUES ("%s")' % (name, value))

        self._sqlite.commit()

    def srem(self, name, *values):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (member text)' % (name,))

        for value in values:
            c.execute('DELETE FROM "%s" WHERE member="%s"' % (name, value))

        self._sqlite.commit()

    def sismember(self, name, value):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (member text)' % (name,))
        c.execute('SELECT * FROM "%s" WHERE member="%s"' % (name, value))

        return c.fetchone() != None

    def smembers(self, name):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (member text)' % (name,))
        c.execute('SELECT * FROM "%s"' % (name,))

        return [row[0] for row in c.fetchall()]

    def hmset(self, name, mapping):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (size text, md5 text)' % (name,))
        c.execute('INSERT INTO "%s" VALUES ("%s", "%s")' % (name, mapping['size'], mapping['md5']))

        self._sqlite.commit()

    def hgetall(self, name):
        c = self._sqlite.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS "%s" (size text, md5 text)' % (name,))
        c.execute('SELECT * FROM "%s"' % (name,))

        row = c.fetchone()

        if row:
            return dict(size=row[0], md5=row[1])

        return dict()



"""MimicDB Bucket subclass wrapper
"""

from boto.s3.bucket import Bucket as BotoBucket
from boto.s3.key import Key as BotoKey

import mimicdb
from .key import Key
from ..backends import tpl


class Bucket(BotoBucket):
    def __init__(self, *args, **kwargs):
        """Set the class for key objects created in the bucket to the MimicDB
        key class.
        """
        kwargs['key_class'] = Key
        super(Bucket, self).__init__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """__iter__ can not be forced to check S3, so return an iterable of
        keys from MimicDB.
        """
        return self.list()

    def get_key(self, *args, **kwargs):
        """Return the key from MimicDB.

        :param boolean force: If true, API call is forced to S3
        """
        if kwargs.pop('force', None):
            headers = kwargs.get('headers', {})
            headers['force'] = True
            kwargs['headers'] = headers

        return super(Bucket, self).get_key(*args, **kwargs)

    def _get_key_internal(self, *args, **kwargs):
        """Return None if key is not in the bucket set.

        Pass 'force' in the headers to check S3 for the key, and after fetching
        the key from S3, save the metadata and key to the bucket set.
        """
        if args[1] is not None and 'force' in args[1]:
            key, res = super(Bucket, self)._get_key_internal(*args, **kwargs)

            if key:
                mimicdb.backend.sadd(tpl.bucket % self.name, key.name)
                mimicdb.backend.hmset(tpl.key % (self.name, key.name),
                                    dict(size=key.size,
                                         md5=key.etag.strip('"')))
            return key, res

        key = None

        if mimicdb.backend.sismember(tpl.bucket % self.name, args[0]):
            key = Key(self)
            key.name = args[0]

        return key, None

    def get_all_keys(self, *args, **kwargs):
        """Return a list of keys from MimicDB.

        :param boolean force: If true, API call is forced to S3
        """
        if kwargs.pop('force', None):
            headers = kwargs.get('headers', args[0] if len(args) else None) or dict()
            headers['force'] = True
            kwargs['headers'] = headers

        return super(Bucket, self).get_all_keys(*args, **kwargs)

    def delete_keys(self, *args, **kwargs):
        """Remove each key or key name in an iterable from the bucket set.
        """
        ikeys = iter(kwargs.get('keys', args[0] if args else []))

        while True:
            try:
                key = ikeys.next()
            except StopIteration:
                break

            if isinstance(key, basestring):
                mimicdb.backend.srem(tpl.bucket % self.name, key)
                mimicdb.backend.delete(tpl.key % (self.name, key))
            elif isinstance(key, BotoKey) or isinstance(key, Key):
                mimicdb.backend.srem(tpl.bucket % self.name, key.name)
                mimicdb.backend.delete(tpl.key % (self.name, key.name))

        return super(Bucket, self).delete_keys(*args, **kwargs)

    def _delete_key_internal(self, *args, **kwargs):
        """Remove key name from bucket set.
        """
        mimicdb.backend.srem(tpl.bucket % self.name, args[0])
        mimicdb.backend.delete(tpl.key % (self.name, args[0]))

        return super(Bucket, self)._delete_key_internal(*args, **kwargs)

    def list(self, *args, **kwargs):
        """Return an iterable of keys from MimicDB.

        :param boolean force: If true, API call is forced to S3
        """
        if kwargs.pop('force', None):
            headers = kwargs.get('headers', args[4] if len(args) > 4 else None) or dict()
            headers['force'] = True
            kwargs['headers'] = headers

            for key in super(Bucket, self).list(*args, **kwargs):
                yield key

        else:
            prefix = kwargs.get('prefix', args[0] if args else '')

            for key in mimicdb.backend.smembers(tpl.bucket % self.name):
                if key.startswith(prefix):
                    k = Key(self, key)

                    meta = mimicdb.backend.hgetall(tpl.key % (self.name, key))

                    if meta:
                        k._load_meta(meta['size'], meta['md5'])

                    yield k

    def _get_all(self, *args, **kwargs):
        """If 'force' is in the headers, retrieve the list of keys from S3.
        Otherwise, use the list() function to retrieve the keys from MimicDB.
        """
        headers = kwargs.get('headers', args[2] if len(args) > 2 else None) or dict()

        if 'force' in headers:
            keys = super(Bucket, self)._get_all(*args, **kwargs)

            for key in keys:
                mimicdb.backend.sadd(tpl.bucket % self.name, key.name)
                mimicdb.backend.hmset(tpl.key % (self.name, key.name), dict(size=key.size, md5=key.etag.strip('"')))

                key.name = key.name

            return keys

        prefix = kwargs.get('prefix', '')

        return list(self.list(prefix=prefix))

    def sync(self):
        """Sync a bucket.

        Force all API calls to S3 and populate the database with the current state of S3.
        """
        for key in mimicdb.backend.smembers(tpl.bucket % self.name):
            mimicdb.backend.delete(tpl.key % (self.name, key))

        mimicdb.backend.delete(tpl.bucket % self.name)
        mimicdb.backend.sadd(tpl.connection, self.name)

        for key in self.list(force=True):
            mimicdb.backend.sadd(tpl.bucket % self.name, key.name)
            mimicdb.backend.hmset(tpl.key % (self.name, key.name), dict(size=key.size, md5=key.etag.strip('"')))



from . import Backend


class Memory(Backend):
    """In-Memory backend. A good example for building a custom backend.

    .. code-block:: python

        from mimicdb.backends.memory import Memory

        memory = Memory()
    """
    def __init__(self):
        self._data = dict()

    def keys(self, pattern='*'):
        pattern = pattern.replace('*', '')
        return [key for key in self._data if key.startswith(pattern)]

    def delete(self, *names):
        for name in names:
            self._data.pop(name, None)

    def sadd(self, name, *values):
        if name in self._data:
            self._data[name].update(values)
        else:
            self._data[name] = set(values)

    def srem(self, name, *values):
        if name in self._data:
            self._data[name].difference_update(values)

    def sismember(self, name, value):
        if name in self._data:
            return value in self._data[name]

        return False

    def smembers(self, name):
        return self._data.get(name, [])

    def hmset(self, name, mapping):
        self._data[name] = mapping

    def hgetall(self, name):
        return self._data.get(name, dict())



"""MimicDB Key subclass wrapper
"""

from boto.s3.key import Key as BotoKey

import mimicdb

from ..backends import tpl


class Key(BotoKey):
    def __init__(self, *args, **kwargs):
        """Add the key to the bucket set if the key name is set and metadata is
        available for it, otherwise wait until uploaded or downloaded.
        """
        bucket = kwargs.get('bucket', args[0] if args else None)
        name = kwargs.get('name', args[1] if len(args) > 1 else None)

        self._name = name

        if name and bucket:
            meta = mimicdb.backend.hgetall(tpl.key % (bucket.name, name))

            if meta:
                mimicdb.backend.sadd(tpl.bucket % bucket.name, name)
                self._load_meta(meta['size'], meta['md5'])

        super(Key, self).__init__(*args, **kwargs)

    def _load_meta(self, size, md5):
        """Set key attributes to retrived metadata. Might be extended in the
        future to support more attributes.
        """
        if not hasattr(self, 'local_hashes'):
            self.local_hashes = {}

        self.size = int(size)
        self.md5 = md5

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        """Key name can be set by Key.key or Key.name. Key.key sets Key.name
        internally, so just handle this property. When changing the key
        name, try to load it's metadata from MimicDB. If it's not available,
        the key hasn't been uploaded, downloaded or synced so don't add it to
        the bucket set (it also might have just been deleted,
        see boto.s3.bucket.py#785)
        """
        self._name = value

        if value:
            meta = mimicdb.backend.hgetall(tpl.key % (self.bucket.name, value))

            if meta:
                mimicdb.backend.sadd(tpl.bucket % self.bucket.name, value)
                self._load_meta(meta['size'], meta['md5'])

    def _send_file_internal(self, *args, **kwargs):
        """Called internally for any type of upload. After upload finishes,
        make sure the key is in the bucket set and save the metadata.
        """
        super(Key, self)._send_file_internal(*args, **kwargs)

        mimicdb.backend.sadd(tpl.bucket % self.bucket.name, self.name)
        mimicdb.backend.hmset(tpl.key % (self.bucket.name, self.name),
                            dict(size=self.size, md5=self.md5))

    def _get_file_internal(self, *args, **kwargs):
        """Called internally for any type of download. After download finishes,
        make sure the key is in the bucket set and save the metadata.
        """
        super(Key, self)._get_file_internal(*args, **kwargs)

        mimicdb.backend.sadd(tpl.bucket % self.bucket.name, self.name)
        mimicdb.backend.hmset(tpl.key % (self.bucket.name, self.name),
                            dict(size=self.size, md5=self.md5))        




from redis import StrictRedis

from . import Backend


class Redis(Backend):
    """Default backend for MimicDB. Initiated with identical parameters
    as ``redis.StrictRedis``.

    :param args \*args, \**kwargs: StrictRedis.__init__() parameters

    .. code-block:: python

        from mimicdb.backends.default import Redis

        redis = Redis(host='localhost', port=6379, db=0)
    """
    def __init__(self, *args, **kwargs):
        self._redis = StrictRedis(*args, **kwargs)

    def keys(self, pattern='*'):
        return self._redis.keys(pattern)

    def delete(self, *names):
        return self._redis.delete(*names)

    def sadd(self, name, *values):
        return self._redis.sadd(name, *values)

    def srem(self, name, *values):
        return self._redis.srem(name, *values)

    def sismember(self, name, value):
        return self._redis.sismember(name, value)

    def smembers(self, name):
        return self._redis.smembers(name)

    def hmset(self, name, mapping):
        return self._redis.hmset(name, mapping)

    def hgetall(self, name):
        return self._redis.hgetall(name)



"""MimicDB S3Connection subclass wrapper
"""

from boto.s3.connection import S3Connection as BotoS3Connection
from boto.exception import S3ResponseError

import mimicdb
from .bucket import Bucket
from ..backends import tpl


class S3Connection(BotoS3Connection):
    def __init__(self, *args, **kwargs):
        """Set the base class for bucket objects created in the connection to
        the MimicDB bucket class.
        """
        kwargs['bucket_class'] = Bucket
        super(S3Connection, self).__init__(*args, **kwargs)

    def get_all_buckets(self, *args, **kwargs):
        """Return a list of buckets in MimicDB.

        :param boolean force: If true, API call is forced to S3
        """
        if kwargs.pop('force', None):
            buckets = super(S3Connection, self).get_all_buckets(*args, **kwargs)

            for bucket in buckets:
                mimicdb.backend.sadd(tpl.connection, bucket.name)

            return buckets

        return [Bucket(self, bucket) for bucket in mimicdb.backend.smembers(tpl.connection)]

    def get_bucket(self, bucket_name, validate=True, headers=None, force=None):
        """Return a bucket from MimicDB if it exists. Return a
        S3ResponseError if the bucket does not exist and validate is passed.

        :param boolean force: If true, API call is forced to S3
        """
        if force:
            bucket = super(S3Connection, self).get_bucket(bucket_name, validate, headers)
            mimicdb.backend.sadd(tpl.connection, bucket.name)
            return bucket

        if mimicdb.backend.sismember(tpl.connection, bucket_name):
            return Bucket(self, bucket_name)
        else:
            if validate:
                raise S3ResponseError(404, 'NoSuchBucket')
            else:
                return Bucket(self, bucket_name)

    def create_bucket(self, *args, **kwargs):
        """Add the bucket to MimicDB after successful creation.
        """
        bucket = super(S3Connection, self).create_bucket(*args, **kwargs)

        if bucket:
            mimicdb.backend.sadd(tpl.connection, bucket.name)

        return bucket

    def delete_bucket(self, *args, **kwargs):
        """Delete the bucket on S3 before removing it from MimicDB.
        If the delete fails (usually because the bucket is not empty), do
        not remove the bucket from the set.
        """
        super(S3Connection, self).delete_bucket(*args, **kwargs)

        bucket = kwargs.get('bucket_name', args[0] if args else None)

        if bucket:
            mimicdb.backend.srem(tpl.connection, bucket)

    def sync(self, *buckets):
        """Sync either a list of buckets or the entire connection.

        Force all API calls to S3 and populate the database with the current
        state of S3.

        :param \*string \*buckets: Buckets to sync
        """
        if buckets:
            for _bucket in buckets:
                for key in mimicdb.backend.smembers(tpl.bucket % _bucket):
                    mimicdb.backend.delete(tpl.key % (_bucket, key))

                mimicdb.backend.delete(tpl.bucket % _bucket)

                bucket = self.get_bucket(_bucket, force=True)

                for key in bucket.list(force=True):
                    mimicdb.backend.sadd(tpl.bucket % bucket.name, key.name)
                    mimicdb.backend.hmset(tpl.key % (bucket.name, key.name), dict(size=key.size, md5=key.etag.strip('"')))
        else:
            for bucket in mimicdb.backend.smembers(tpl.connection):
                for key in mimicdb.backend.smembers(tpl.bucket % bucket):
                    mimicdb.backend.delete(tpl.key % (bucket, key))

                mimicdb.backend.delete(tpl.bucket % bucket)

            for bucket in self.get_all_buckets(force=True):
                for key in bucket.list(force=True):
                    mimicdb.backend.sadd(tpl.bucket % bucket.name, key.name)
                    mimicdb.backend.hmset(tpl.key % (bucket.name, key.name), dict(size=key.size, md5=key.etag.strip('"')))



from setuptools import setup, find_packages

setup(
    name = 'MimicDB',
    version = '1.0.3',
    packages = find_packages(),
    install_requires = ['boto>=2.32.1',],
    license = 'BSD',
    author = 'Nathan Cahill',
    author_email = 'nathan@nathancahill.com',
    url = 'https://github.com/nathancahill/mimicdb',
    description='Python implementation of MimicDB',
    long_description=open('README.txt').read(),
)



class Backend(object):
    """Base class for MimicDB backends. Extendable to support custom backends.
    A custom backend **must** implement each of the functions of the
    base class.

    .. code-block:: python

        from mimicdb.backends import Backend

        class MyAwesomeBackend(Backend):
            def __init__(self):

            etc.
    """
    def __init__(self):
        pass

    def delete(self, *names):
        pass

    def sadd(self, name, *values):
        pass

    def srem(self, name, *values):
        pass

    def sismember(self, name, value):
        pass

    def smembers(self, name):
        pass

    def hmset(self, name, mapping):
        pass

    def hgetall(self, name):
        pass
